import sys

def getLibHeaderFilesPath ():
    """
    """
    path = ''
    try:
        import subprocess
        path = subprocess.check_output (['pkg-config', '--cflags-only-I', 'libunity_webapps-0.2']).strip()
    except:
        pass
    
    if len(path) != 0 and path.startswith ('-I'):
        path = path[len("-I"):]
		
    return path

def getHeaderFilesFrom (path):
    """
    """de
    import os
    return [file for file in os.listdir (path) if file.endswith ('.h')]


def getBindingFunctionPrefix ():
    return "unity_webapps_binding_"


def getUnityWebappsFunctionPrefix ():
    return "unity_webapps_"


def getCandidatesForHeaderFileGen (files):
    """
    Takes a list of files (header files) and returns a list of files
    that should  be considered for "webapps binding" generation
    """
    candidates = [file for file in files if file.endswith('-context.h') and -1 == file.find('-gen-') or file.endswith('webapps-service.h') or file.endswith('script-repo.h')]
    return candidates


def getFunctionDefinitionsFrom (filename):
    """
    """
    lines = open(filename).read().split(';')
    lines = [line for line in lines if not line.strip().startswith('typedef')]
    import re
    fun_re = re.compile (r'(([^\(\)\*]*)(\s*\*\s*|\s+))(\w+)\s*\(([^\(\)]*)\)\s*')
    defs = []
    for line in lines:
        m = fun_re.match (line)
        if m and len(m.groups()) >= 3:
            defs.append({"name": m.groups()[3].strip(), "return_type": m.groups()[0].strip(), "args": [param.strip() for param in m.groups()[4].split(',')]})
    return defs


def getSourceFileBodyTemplate ():
    return """
/*
 * unity-webapps-binding.c
 * Copyright (C) Canonical LTD 2012
 * 
 * Author: Alexandre Abreu <alexandre.abreu@canonical.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#include <glib.h>

// TODO change that, proper path
#include <libunity_webapps-0.2/unity-webapps-service.h>
#include <libunity_webapps-0.2/unity-webapps-context.h>
#include <libunity_webapps-0.2/unity-webapps-music-player-context.h>
#include <libunity_webapps-0.2/unity-webapps-permissions.h>
#include <libunity_webapps-0.2/unity-webapps-launcher-context.h>
#include <libunity_webapps-0.2/unity-webapps-notification-context.h>

#include "npapi-headers/headers/npapi.h"
#include "npapi-headers/headers/npfunctions.h"
#include "npapi-headers/headers/npruntime.h"

#include "wrapped-ptr-type.h"
#include "wrapped-callback-type.h"

#include "unity-webapps-scriptable-object.h"
#include "unity-webapps-repo-userscript-wrapper.h"

static bool is_null_or_void (NPVariant variant)
{
  return NPVARIANT_IS_NULL(variant) || NPVARIANT_IS_VOID(variant);
}

static const gchar * create_safe_string_for (const NPVariant * const pVariant)
{
  g_return_val_if_fail (NULL != pVariant, NULL);

  char * pSafeString = (gchar*)g_malloc0 (NPVARIANT_TO_STRING(*pVariant).UTF8Length + 1);
  g_strlcpy (pSafeString
	     , NPVARIANT_TO_STRING(*pVariant).UTF8Characters
	     , NPVARIANT_TO_STRING(*pVariant).UTF8Length + 1);

  return (const gchar *) pSafeString;
}

%s
"""


def getHeaderBodyTemplate ():
    return """
/*
 * unity-webapps-binding.h
 * Copyright (C) Canonical LTD 2012
 * 
 * Author: Alexandre Abreu <alexandre.abreu@canonical.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#if ! defined (UNITY_WEBAPPS_BINDING_H)
#define UNITY_WEBAPPS_BINDING_H

#include "npapi-headers/headers/npapi.h"
#include "npapi-headers/headers/nptypes.h"

%s

#endif // UNITY_WEBAPPS_BINDING_H

"""



##
##
## binding declarations/header file generation
##
## 

def generateBindingFunctionDeclFor (functioname):
    return """
NPVariant
%s%s (NPP instance
         , NPObject * npobject
         , const NPVariant *args
         , uint32_t argCount);
""" % (getBindingFunctionPrefix (), functioname)


def generateBindingHeaderFileForFunctions (functions, headerFilename):
    functionBindingNames = [function['name'].replace (getUnityWebappsFunctionPrefix(), '') for function in functions]
    
    decls = ''
    for functionBinding in functionBindingNames:
        decls += generateBindingFunctionDeclFor (functionBinding) + "\n"
    return decls


def getCallbackDefinitionsFrom (file):
    lines = open(file).read().split(';')
    import re
    cb_re = re.compile (r'\s*typedef\s*\w*\s*\(\s*\*\s*\w*\s*\)\s*\(.*\)\s*')
    cbs = []
    for line in lines:
        if cb_re.match (line):
            cbs.append(line.strip())
    return cbs


def generateBindingHeaderFile (candidateHeaderFiles, headerFilename):
    functions = []
    for candidate in candidateHeaderFiles:
        functions.extend (getFunctionDefinitionsFrom (candidate))
    
    decls = '// This file has been generated'
    decls += getHeaderBodyTemplate () % generateBindingHeaderFileForFunctions (functions, headerFilename)
    
    open(headerFilename, 'w+').write (decls)



##
##
## binding definition/source file generation
##
## 

def generateFunctionArgumentCountCheck (count):
    return """
    if (argCount != %s)
    {
	// TODO find better error reporting mecanism (exceptions?)
	return result;
    }
    """ % str(count)

def extractArgumentNamesFromArgs (args):
    """
        Extract the parameter names from a given list of C function arguments
    
        @param args list of string (C argument literals)
        @return list of extracted parameter names
    """
    import re
    cargre = re.compile (r'^.+[\*|\s+]([^\s*]+)\s*$')
    xargs = [''] * len(args)
    return map (lambda l: cargre.match (l[1]) and cargre.match (l[1]).groups()[0].strip() or args[l[0]], enumerate(args))


def extractTypeNamesFromArgs (args):
    """
        Extract the parameter types from a given list of C function arguments
    
        @param args list of string (C argument literals)
        @return list of extracted arguments
    """
    import re
    cargre = re.compile (r'^([\s\w_]*[\*\s])\D[\w_]+\s*$')
    xargs = [''] * len(args)
    return map (lambda l: cargre.match (l[1]) and cargre.match (l[1]).groups()[0].strip() or args[l[0]], enumerate(args))


def stripType (t):
    qualifiers = "const"    
    return t.strip ().replace (qualifiers, "").replace (" ", "")


def getInfoForType (t):
    baseTypes = {"gchar*":
                 {"validator": "NPVARIANT_IS_STRING", "can_free": True, 'create_var': True, 'creator': 'create_safe_string_for (&args[%d])', 'null': 'NULL', 'nullable': True, 'to_variant': 'STRINGZ_TO_VARIANT'}
                 , "gint":
                 {"validator": "NPVARIANT_IS_INT32", "can_free": False, 'create_var': True, 'creator': 'NPVARIANT_TO_INT32 (args[%d])', 'null': '0', 'nullable': False, 'to_variant': 'INT32_TO_NPVARIANT'}
                 , "guint64":
                 {"validator": "NPVARIANT_IS_DOUBLE", "can_free": False, 'create_var': True, 'creator': 'NPVARIANT_TO_DOUBLE (args[%d])', 'null': '0', 'nullable': False, 'to_variant': 'INT32_TO_NPVARIANT'}
                 , "guint":
                 {"validator": "NPVARIANT_IS_INT32", "can_free": False, 'create_var': True, 'creator': 'NPVARIANT_TO_INT32 (args[%d])', 'null': '0', 'nullable': False, 'to_variant': 'INT32_TO_NPVARIANT'}
                 , "gboolean":
                 {"validator": "NPVARIANT_IS_BOOLEAN", "can_free": False, 'create_var': True, 'creator': 'NPVARIANT_TO_BOOLEAN (args[%d])', 'null': 'false', 'nullable': False, 'to_variant': 'BOOLEAN_TO_NPVARIANT'}
                 , "gsize":
                 {"validator": "NPVARIANT_IS_INT32", "can_free": False, 'create_var': True,  'creator': 'NPVARIANT_TO_INT32 (args[%d])', 'null': '0', 'nullable': False, 'to_variant': 'INT32_TO_NPVARIANT'}
                 , "gdouble":
                 {"validator": "NPVARIANT_IS_DOUBLE", "can_free": False, 'create_var': True, 'creator': 'NPVARIANT_TO_BOOLEAN (args[%d])', 'null': '0.0', 'nullable': False, 'to_variant': 'DOUBLE_TO_VARIANT'}
                 , "void":
                 {"validator": "NPVARIANT_IS_VOID", "can_free": False, 'create_var': False, 'creator': None, 'null': None, 'nullable': False, 'to_variant': 'VOID_TO_NPVARIANT'}
                 , "gpointer":
                 {"validator": "NPVARIANT_IS_OBJECT", "can_free": False, 'create_var': True, 'creator': None, 'null': 'NULL', 'nullable': True, 'to_variant': 'OBJECT_TO_NPVARIANT'}
             }
    
    # defaults to object
    other = {'validator': 'NPVARIANT_IS_OBJECT', 'can_free': False, 'create_var': True, 'creator': '(void*) ((wrapped_void_ptr_t *) args[%d].value.objectValue)->pWrapped', 'null': 'NULL', 'nullable': True, 'to_variant': 'OBJECT_TO_NPVARIANT'}
    
    t = stripType(t).strip()
    
    info = baseTypes.has_key (t) and baseTypes[t] or other
    print t, info['validator']
    return info


def generateFunctionArgumentTypeValidation (args):
    validators = []
    types = extractTypeNamesFromArgs (args)
    for i, t in enumerate(types):
        arg = '(args[' + str(i) + '])'
        validators.append (' ( ! ' + getInfoForType (t)['validator'] + arg + ' && ! is_null_or_void' + arg + ' )')
    
    if len(validators) == 0:
        return ""
    
    return """
    if ( %s )
    {
        // TODO better error reporting
        return result;
    }
    """ % " || \n\t".join (validators)


def generateVariableCreationForArg (idx, arg_type, arg_name):
    infos = getInfoForType (arg_type)
    src = ''
    if infos['create_var']:
        src += "%s %s = %s;\n" % (arg_type, arg_name, infos['null'])
        if not infos['creator'] is None:
            src += """
            if ( ! is_null_or_void (args[%d]))
            {
            \t%s = %s;
            }
            """ % (idx, arg_name, infos['creator'] % idx)
    return src


def generateVariableCreationForArgs (args):
    arg_infos = zip (range(len(args)), extractTypeNamesFromArgs(args), extractArgumentNamesFromArgs(args))
    gen = []
    
    for idx, arg_type, arg_name in arg_infos:
        gen.append (generateVariableCreationForArg (idx, arg_type, arg_name))
        
    return ';\n'.join (gen)


def generateWebappsFunctionCallFor (func):
    args = extractArgumentNamesFromArgs (func['args'])
    name = getUnityWebappsFunctionPrefix () + func['name']
    
    func_call = name + ' (' + ', '.join (args) + ');'
    
    if not func['return_type'].strip ().startswith ('void'):
        func_call = func['return_type'] + ' ret_val = ' + func_call
        
        if getInfoForType (func['return_type'])['nullable']:
            func_call = func_call + """
            if (NULL == ret_val)
            {
                // TODO better error reporting
                return result;
            }
            
            NPObject * object = create_wrapped_ptr_object_for (instance, ret_val);
            if (NULL == object)
            {
                // TODO find better error reporting mecanism (exceptions?)
                return result;
            }
            
            OBJECT_TO_NPVARIANT(object, result);
            
            """
        else:
            func_call = func_call + """
            
            %s (ret_val, result);
            
            """ % getInfoForType (func['return_type'])['to_variant']
    
    return func_call


def generateBindingFunctionDefinitionFor (func_description):
    function_name = func_description['name'].replace (getUnityWebappsFunctionPrefix (), '')
    print "******* ", function_name
    return """
NPVariant
    %s%s (NPP instance
                            , const NPVariant *args
                            , uint32_t argCount)
{
    NPVariant result;
    NULL_TO_NPVARIANT (result);
    
    %s
    
    %s
    
    %s
    

//    g_free (name);
//    g_free (domain);
//    g_free (icon_url);

    %s

    return result;
}
    """ % (getBindingFunctionPrefix ()
           , function_name
        , generateFunctionArgumentCountCheck(len(func_description['args']))
        , generateFunctionArgumentTypeValidation(func_description['args'])
        , generateVariableCreationForArgs(func_description['args'])
        , generateWebappsFunctionCallFor (func_description)
        )


def generateBindingSourceFileForFunctions (functions, headerFilename):
    decls = ''
    for function in functions:
        decls += generateBindingFunctionDefinitionFor (function) + "\n"
    
    return decls


def generateBindingSourceFile (candidateHeaderFiles, headerFilename):
    functions = []
    for candidate in candidateHeaderFiles:
        functions += getFunctionDefinitionsFrom (candidate)
    
    decls = '// This file has been generated'
    decls += getSourceFileBodyTemplate () % generateBindingSourceFileForFunctions (functions, headerFilename)
    
    open(headerFilename, 'w+').write (decls)



if __name__ == "__main__":
    import sys
    
    filename = 'unity-webapps-binding-test'
#    if len(sys.argv) != 2:
#        sys.exit(1)
    
    import os
    
    headerpath = getLibHeaderFilesPath ()
    candidates = [os.path.join (headerpath, candidate) for candidate in getCandidatesForHeaderFileGen (getHeaderFilesFrom (headerpath))]
    
    generateBindingHeaderFile (candidates, filename + '.h')
    generateBindingSourceFile (candidates, filename + '.c')
    
